unit HintOpen;
{The component display usual FileOpen dialog box. If it was displayed
in Explorer style, new properties - OnChange is added. The procedure will
be called each time:
  1. If one has changed selection of a file
  2. If one has changed filter selection
Second property - OnPaint is used to fill hint window arisen. If the property
was assigned, then some hint is shown. One can insert in it content of
selected file.
Using the property one can read file with specified filter and to display
information on the file selected as hint

FreeWare. No warranties. If one will distribute the source code, reference on
author (below) is required. If one will modify the source code, highlighting
(by any means) of modifications is required.

Author: Sergei Trepalin
E-mail: trep@trep.home.chg.ru}

{$R-}

interface

uses Windows, Messages, SysUtils, CommDlg, Classes, Graphics, Controls,
  Forms, Dialogs, HintWnd;

type
 {TOpenDialogHint }
  TOpenDialogHint = class(TOpenDialog)
  private
    FGraphicHint:TGraphHintWindow;
    FOldFileName:string;
    FHintSize:word;
    FOldFilterIndex:integer;
    FFileChange:TNotifyEvent;
    FHintDraw:TGraphicDrawNotify;
    function NewExecute(Func: Pointer): Bool;
  public
    FFilterSelected:boolean; {Internal flag, correct bug in standard
    TFileOpen class - sending notification about file selection when
    filter selection is perfomed}
    constructor Create(AOwner:TComponent); override;
    function Execute: Boolean; override;
    property SelectedFileName:string read FOldFilename;
    property SelectedFilter:integer read FOldFilterIndex;
  published
    property OnChange:TNotifyEvent read FFileChange write FFileChange;
    property OnPaint:TGraphicDrawNotify read FHintDraw write FHintDraw;
    property HintSize:word read FHintSize write FHintSize default 150;
  end;

procedure Register;

implementation
var
  HookCtl3D: Boolean;
  OpenDialogID:TOpenDialogHint;

{1:1 function from Dialogs unit}
procedure CenterWindow(Wnd: HWnd);
var
  Rect: TRect;
begin
  GetWindowRect(Wnd, Rect);
  SetWindowPos(Wnd, 0,
    (GetSystemMetrics(SM_CXSCREEN) - Rect.Right + Rect.Left) div 2,
    (GetSystemMetrics(SM_CYSCREEN) - Rect.Bottom + Rect.Top) div 3,
    0, 0, SWP_NOACTIVATE or SWP_NOSIZE or SWP_NOZORDER);
end;

{1:1 function from Dialogs unit. Old dialog style can not support
 identification of selections during run-time}
function DialogHook(Wnd: HWnd; Msg: UINT; WParam: WPARAM; LParam: LPARAM): UINT; stdcall;
begin
  Result := 0;
  case Msg of
    WM_INITDIALOG:
      begin
        if HookCtl3D then
        begin
          Subclass3DDlg(Wnd, CTL3D_ALL);
          SetAutoSubClass(True);
        end;
        CenterWindow(Wnd);
        Result := 1;
      end;
    WM_DESTROY:
      if HookCtl3D then SetAutoSubClass(False);
  end;
end;

{The function was greatly overrided}
function ExplorerHook(Wnd: HWnd; Msg: UINT; WParam: WPARAM; LParam: LPARAM): UINT; stdcall;
var
  FP:array[0..255] of char;
  S:string;
  I,N:integer;
  F:file;
  PC:TPoint;
  R:TRect;
  TestNew:boolean;
begin
  Result := 0;
  TestNew:=False;
  if Msg=WM_NOTIFY then begin
    if (POFNotify(LParam)^.hdr.code = CDN_INITDONE) then
     CenterWindow(GetWindowLong(Wnd, GWL_HWNDPARENT));
    {My part below}
    if (POFNotify(LParam)^.hdr.code=CDN_SELCHANGE) then begin
      SendMessage(POFNotify(LParam)^.hdr.hwndFrom,CDM_GETFILEPATH,255,longint(@FP[0]));
      S:=StrPas(FP);
      N:=POFNotify(LParam)^.lpOFN^.nFilterIndex;
      {$I-}
      if ((S<>OpenDialogID.FOldFilename) or (N<>OpenDialogID.FOldFilterIndex))
       and (not OpenDialogID.FFilterSelected) then begin
        {Checking that filename was selected - not folder!}
        AssignFile(F,S);
        Reset(F,1);
        I:=IOResult;
        if I=0 then begin
          {File exists! - hurrah!}
          Close(F);
          OpenDialogID.FOldFileName:=S;
          OpenDialogID.FOldFilterIndex:=N;
          {Notification about selection change}
          if Assigned(OpenDialogID.FFileChange) then OpenDialogID.FFileChange(OpenDialogID);
          {And then paint - if assigned}
          if Assigned(OpenDialogID.FHintDraw) then begin
            if Assigned(OpenDialogID.FGraphicHint) then DestroyWindow(OpenDialogID.FGraphicHint.Handle);
            OpenDialogID.FGraphicHint:=nil;
            try
              OpenDialogID.FGraphicHint:=TGraphHintWindow.Create(OpenDialogID);
              OpenDialogID.FGraphicHint.OnPaint:=OpenDialogID.OnPaint;
            except
              Raise;
              OpenDialogID.FGraphicHint:=nil;
            end;
            if Assigned(OpenDialogID.FGraphicHint) then begin
              GetCursorPos(PC);
              R.Left:=PC.X-(OpenDialogID.HintSize div 2);
              R.Top:=PC.Y+40;
              R.Right:=R.Left+OpenDialogID.HintSize;
              R.Bottom:=R.Top+OpenDialogID.HintSize;
              OpenDialogID.FGraphicHint.ActivateHint(R);
              TestNew:=True;
            end;
          end;
        end;
      end;
      OpenDialogID.FFilterSelected:=False;
    end else  if (POFNotify(LParam)^.hdr.code=CDN_TYPECHANGE) then OpenDialogID.FFilterSelected:=True;
  end;
  {Disposing hint window in another kind of messages}
  if (not TestNew) and Assigned(OpenDialogID.FGraphicHint) then begin
    DestroyWindow(OpenDialogID.FGraphicHint.Handle);
    OpenDialogID.FGraphicHint:=nil;
  end;
  {Message CDN_TYPECHANGE will not be covered because of changing
  of filter index  will generate CDN_SELCHANGE authomatically}
end;

{ TOpenDialog }
constructor TOpenDialogHint.Create;
begin
  inherited Create(AOwner);
  {The only purpose to override Create method is to store reference on
  created dialog as global variable. The variable will use in hook function
  which I can not to determine as function of object.
      Obviously, if one creates two dialogs at time in a single instance,
  the unit will fail}
  OpenDialogID:=Self;
  FOldFileName:='';
  FOldFilterIndex:=-1;
  FFileChange:=nil;
  FHintDraw:=nil;
  FGraphicHint:=nil;
  FHintSize:=150;
  FFilterSelected:=False;
end;

function TOpenDialogHint.NewExecute(Func: Pointer): Bool;
{This function was copied 1:1 from Dialogs unit, except that new hook
are called from it. Look in Dialogs unit for explanations}
const
  MultiSelectBufferSize = 8192;
  OpenOptions: array [TOpenOption] of Longint = (
    OFN_READONLY, OFN_OVERWRITEPROMPT, OFN_HIDEREADONLY,
    OFN_NOCHANGEDIR, OFN_SHOWHELP, OFN_NOVALIDATE, OFN_ALLOWMULTISELECT,
    OFN_EXTENSIONDIFFERENT, OFN_PATHMUSTEXIST, OFN_FILEMUSTEXIST,
    OFN_CREATEPROMPT, OFN_SHAREAWARE, OFN_NOREADONLYRETURN,
    OFN_NOTESTFILECREATE, OFN_NONETWORKBUTTON, OFN_NOLONGNAMES,
    OFN_EXPLORER, OFN_NODEREFERENCELINKS);
var
  Option: TOpenOption;
  P: PChar;
  CDefaultExt: array[0..3] of Char;
  OpenFilename: TOpenFilename;
  S:TFileName;
  MyOptions:TOpenOptions;

  function AllocFilterStr(const S: string): PChar;
  var
    P: PChar;
  begin
    Result := nil;
    if S <> '' then
    begin
      Result := StrCopy(StrAlloc(Length(S) + 2), PChar(S));
      P := Result;
      while P^ <> #0 do
      begin
        if P^ = '|' then P^ := #0;
        Inc(P);
      end;
      Inc(P);
      P^ := #0;
    end;
  end;

  function FindExtension(P: PChar): PChar;
  begin
    Result := '';
    while P^ <> #0 do
    begin
      if P^ = '.' then Result := P + 1 else
        if P^ = '\' then Result := '';
      Inc(P);
    end;
  end;

  function ExtractFileName(P: PChar; var S: string): PChar;
  var
    Separator: Char;
  begin
    Separator := #0;
    if (ofAllowMultiSelect in Options) and
      ((ofOldStyleDialog in Options) or not NewStyleControls) then
      Separator := ' ';
    Result := P;
    while (Result[0] <> #0) and (Result[0] <> Separator) do Inc(Result);
    SetString(S, P, Result - P);
    if Result[0] = Separator then Inc(Result);
  end;

  procedure ExtractFileNames(P: PChar);
  var
    DirName, FileName: string;
  begin
    P := ExtractFileName(P, DirName);
    P := ExtractFileName(P, FileName);
    if FileName = '' then
      Files.Add(DirName)
    else
    begin
      if DirName[Length(DirName)] <> '\' then
        DirName := DirName + '\';
      repeat
        if (FileName[1] <> '\') and ((Length(FileName) <= 3) or
          (FileName[2] <> ':') or (FileName[3] <> '\')) then
          FileName := DirName + FileName;
        Files.Add(FileName);
        P := ExtractFileName(P, FileName);
      until FileName = '';
    end;
  end;

begin
  Files.Clear;
  FillChar(OpenFileName, SizeOf(OpenFileName), 0);
  with OpenFilename do
  try
    lStructSize := SizeOf(TOpenFilename);
    hInstance := {Application.Handle}0;
    lpstrFilter := AllocFilterStr(Filter);
    nFilterIndex := FilterIndex;
    if ofAllowMultiSelect in Options then
      nMaxFile := MultiSelectBufferSize else
      nMaxFile := MAX_PATH;
    GetMem(lpstrFile, nMaxFile + 2);
    FillChar(lpstrFile^, nMaxFile + 2, 0);
    StrLCopy(lpstrFile, PChar(FileName), nMaxFile);
    lpstrInitialDir := PChar(InitialDir);
    lpstrTitle := PChar(Title);
    HookCtl3D := Ctl3D;
    Flags := OFN_ENABLEHOOK;
    for Option := Low(Option) to High(Option) do
      if Option in Options then
        Flags := Flags or OpenOptions[Option];
    if NewStyleControls then
      Flags := Flags xor OFN_EXPLORER
    else
      Flags := Flags and not OFN_EXPLORER;
    if DefaultExt <> '' then
    begin
      P := PChar(DefaultExt);
      if (P^ = #0) and (Flags and OFN_EXPLORER = 0) then
        P := FindExtension(PChar(FileName));
      lpstrDefExt := StrLCopy(CDefaultExt, P, 3)
    end;
    if (ofOldStyleDialog in Options) or not NewStyleControls then
      {New function - defined by me - S.T.}
      lpfnHook := DialogHook
    else
      {New function - defined by me - S.T.}
      lpfnHook := ExplorerHook;
    hWndOwner := Application.Handle;
    Result := TaskModalDialog(Func, OpenFileName);
    if Result then
    begin
      if ofAllowMultiSelect in Options then
      begin
        ExtractFileNames(lpstrFile);
        FileName := Files[0];
      end else
      begin
        ExtractFileName(lpstrFile,S{ FFileName});
        FileName:=S;
        Files.Add(FileName);
      end;
      MyOptions:=Options;
      if (Flags and OFN_EXTENSIONDIFFERENT) <> 0 then
        Include(MyOptions, ofExtensionDifferent) else
        Exclude(MyOptions, ofExtensionDifferent);
      if (Flags and OFN_READONLY) <> 0 then
        Include(MyOptions, ofReadOnly) else
        Exclude(MyOptions, ofReadOnly);
      Options:=MyOptions;
      FilterIndex := nFilterIndex;
    end;
  finally
    if lpstrFile <> nil then FreeMem(lpstrFile, nMaxFile + 2);
    if lpstrFilter <> nil then StrDispose(lpstrFilter);
  end;
end;

function TOpenDialogHint.Execute: Boolean;
begin
 {Purpose to override the function - to call new function with new
  hooks. One can wonder - very large amount of source codes was
  copied 1:1 from Dialogs unit. Please, give Your alternate solution...}
  Result := NewExecute(@GetOpenFileName);
end;

procedure Register;
begin
  RegisterComponents('Interface', [TOpenDialogHint]);
end;

end.
